@@ -289,17 +289,6 @@ free_param_data(SynapticsPrivate *priv)
     priv->synpara = NULL;
 }
 
-static double
-synSetFloatOption(pointer options, const char *optname, double default_value)
-{
-    char *str_par;
-    double value;
-    str_par = xf86FindOptionValue(options, optname);
-    if ((!str_par) || (sscanf(str_par, "%lf", &value) != 1))
-	return default_value;
-    return value;
-}
-
 static void set_default_parameters(LocalDevicePtr local)
 {
     SynapticsPrivate *priv = local->private; /* read-only */
@@ -313,6 +302,7 @@ static void set_default_parameters(LocalDevicePtr local)
     double accelFactor;					/* 1/pixels */
     int fingerLow, fingerHigh, fingerPress;		/* pressure */
     int emulateTwoFingerMinZ;				/* pressure */
+    int emulateTwoFingerMinW;				/* width */
     int edgeMotionMinZ, edgeMotionMaxZ;			/* pressure */
     int pressureMotionMinZ, pressureMotionMaxZ;		/* pressure */
     int palmMinWidth, palmMinZ;				/* pressure */
@@ -372,15 +362,15 @@ static void set_default_parameters(LocalDevicePtr local)
 	int range = priv->maxp - priv->minp;
 
 	/* scaling based on defaults below and a pressure of 256 */
-	fingerLow = priv->minp + range * .098;
-	fingerHigh = priv->minp + range * .117;
+	fingerLow = priv->minp + range * (25.0/256);
+	fingerHigh = priv->minp + range * (30.0/256);
 	fingerPress = priv->minp + range * 1.000;
-	emulateTwoFingerMinZ = priv->minp + range * 1.1;
-	edgeMotionMinZ = priv->minp + range * .117;
-	edgeMotionMaxZ = priv->minp + range * .625;
-	pressureMotionMinZ = priv->minp + range * .117;
-	pressureMotionMaxZ = priv->minp + range * .625;
-	palmMinZ = priv->minp + range * .781;
+	emulateTwoFingerMinZ = priv->minp + range * (282.0/256);
+	edgeMotionMinZ = priv->minp + range * (30.0/256);
+	edgeMotionMaxZ = priv->minp + range * (160.0/256);
+	pressureMotionMinZ = priv->minp + range * (30.0/256);
+	pressureMotionMaxZ = priv->minp + range * (160.0/256);
+	palmMinZ = priv->minp + range * (200.0/256);
     } else {
 	fingerLow = 25;
 	fingerHigh = 30;
@@ -397,10 +387,11 @@ static void set_default_parameters(LocalDevicePtr local)
 	int range = priv->maxw - priv->minw;
 
 	/* scaling based on defaults below and a tool width of 16 */
-	palmMinWidth = priv->minw + range * .625;
-
+	palmMinWidth = priv->minw + range * (10.0/16);
+	emulateTwoFingerMinW = priv->minw + range * (7.0/16);
     } else {
 	palmMinWidth = 10;
+	emulateTwoFingerMinW = 7;
     }
 
     /* Enable tap if we don't have a phys left button */
@@ -438,10 +429,15 @@ static void set_default_parameters(LocalDevicePtr local)
     pars->fast_taps = xf86SetBoolOption(opts, "FastTaps", FALSE);
     pars->emulate_mid_button_time = xf86SetIntOption(opts, "EmulateMidButtonTime", 75);
     pars->emulate_twofinger_z = xf86SetIntOption(opts, "EmulateTwoFingerMinZ", emulateTwoFingerMinZ);
+    pars->emulate_twofinger_w = xf86SetIntOption(opts, "EmulateTwoFingerMinW", emulateTwoFingerMinW);
     pars->scroll_dist_vert = xf86SetIntOption(opts, "VertScrollDelta", horizScrollDelta);
     pars->scroll_dist_horiz = xf86SetIntOption(opts, "HorizScrollDelta", vertScrollDelta);
     pars->scroll_edge_vert = xf86SetBoolOption(opts, "VertEdgeScroll", vertEdgeScroll);
-    pars->special_scroll_area_right  = xf86SetBoolOption(opts, "SpecialScrollAreaRight", TRUE);
+    if (xf86CheckIfOptionUsedByName(opts, "RightEdge")) {
+      pars->special_scroll_area_right  = FALSE;
+    } else {
+      pars->special_scroll_area_right  = xf86SetBoolOption(opts, "SpecialScrollAreaRight", TRUE);
+    }
     pars->scroll_edge_horiz = xf86SetBoolOption(opts, "HorizEdgeScroll", horizEdgeScroll);
     pars->scroll_edge_corner = xf86SetBoolOption(opts, "CornerCoasting", FALSE);
     pars->scroll_twofinger_vert = xf86SetBoolOption(opts, "VertTwoFingerScroll", vertTwoFingerScroll);
@@ -480,14 +476,14 @@ static void set_default_parameters(LocalDevicePtr local)
     pars->press_motion_min_z = xf86SetIntOption(opts, "PressureMotionMinZ", pressureMotionMinZ);
     pars->press_motion_max_z = xf86SetIntOption(opts, "PressureMotionMaxZ", pressureMotionMaxZ);
 
-    pars->min_speed = synSetFloatOption(opts, "MinSpeed", 0.4);
-    pars->max_speed = synSetFloatOption(opts, "MaxSpeed", 0.7);
-    pars->accl = synSetFloatOption(opts, "AccelFactor", accelFactor);
-    pars->trackstick_speed = synSetFloatOption(opts, "TrackstickSpeed", 40);
-    pars->scroll_dist_circ = synSetFloatOption(opts, "CircScrollDelta", 0.1);
-    pars->coasting_speed = synSetFloatOption(opts, "CoastingSpeed", 0.0);
-    pars->press_motion_min_factor = synSetFloatOption(opts, "PressureMotionMinFactor", 1.0);
-    pars->press_motion_max_factor = synSetFloatOption(opts, "PressureMotionMaxFactor", 1.0);
+    pars->min_speed = xf86SetRealOption(opts, "MinSpeed", 0.4);
+    pars->max_speed = xf86SetRealOption(opts, "MaxSpeed", 0.7);
+    pars->accl = xf86SetRealOption(opts, "AccelFactor", accelFactor);
+    pars->trackstick_speed = xf86SetRealOption(opts, "TrackstickSpeed", 40);
+    pars->scroll_dist_circ = xf86SetRealOption(opts, "CircScrollDelta", 0.1);
+    pars->coasting_speed = xf86SetRealOption(opts, "CoastingSpeed", 0.0);
+    pars->press_motion_min_factor = xf86SetRealOption(opts, "PressureMotionMinFactor", 1.0);
+    pars->press_motion_max_factor = xf86SetRealOption(opts, "PressureMotionMaxFactor", 1.0);
     pars->grab_event_device = xf86SetBoolOption(opts, "GrabEventDevice", TRUE);
 
     /* Warn about (and fix) incorrectly configured TopEdge/BottomEdge parameters */
@@ -591,7 +587,7 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     }
 
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
-    local->history_size = xf86SetIntOption(opts, "HistorySize", 0);
+    local->history_size = xf86SetIntOption(local->options, "HistorySize", 0);
 #endif
 
     xf86ProcessCommonOptions(local, local->options);
@@ -891,7 +887,7 @@ edge_detection(SynapticsPrivate *priv, int x, int y)
     if (priv->synpara->circular_pad)
 	return circular_edge_detection(priv, x, y);
 
-    if (x >= priv->synpara->right_edge)
+    if (x > priv->synpara->right_edge)
 	edge |= RIGHT_EDGE;
     else if (x < priv->synpara->left_edge)
 	edge |= LEFT_EDGE;
@@ -1938,7 +1934,8 @@ HandleState(LocalDevicePtr local, struct SynapticsHwState *hw)
     }
 
     /* Two finger emulation */
-    if (hw->z >= para->emulate_twofinger_z && hw->numFingers == 1) {
+    if (hw->numFingers == 1 && hw->z >= para->emulate_twofinger_z &&
+        hw->fingerWidth >= para->emulate_twofinger_w) {
 	hw->numFingers = 2;
     }
 
@@ -1970,32 +1967,20 @@ HandleState(LocalDevicePtr local, struct SynapticsHwState *hw)
 
     /*
      * Some touchpads have a scroll wheel region where a very large X
-     * coordinate is reported. 
-     *
-     *    We suggest two  solution this problem:
+     * coordinate is reported. In this case for eliminate discontinuity,
+     * we adjust X and simulate new zone which adjacent to right edge.
      */
     if (hw->x <= XMAX_VALID) {
 	if (priv->largest_valid_x < hw->x)
 	    priv->largest_valid_x = hw->x;
     } else {
-      if (!(para->special_scroll_area_right))
-      /*               First:
-      * Adjust the X coordinate to eliminate the discontinuity 
-      * and use it region as 1 coordinate size line.
-      */
 	hw->x = priv->largest_valid_x + 1;
-      else {
-      /*               Second (default):
-       * Adjust the X coordinate to eliminate the discontinuity
-       * and use it region as scroll area automaticly.
-       */	
-	
-	if (priv->synpara->right_edge > priv->largest_valid_x + 1)
-	  priv->synpara->right_edge=priv->largest_valid_x + 1;
-	para->special_scroll_area_right = FALSE;
-       
-	hw->x = priv->largest_valid_x + 1;
-      }
+    /*
+     * If user didn't set right_edge manualy, auto-adjust to bounds of
+     * hardware scroll area.
+     */
+	if (para->special_scroll_area_right)
+	  priv->synpara->right_edge = priv->largest_valid_x;
     }
 
     edge = edge_detection(priv, hw->x, hw->y);
